#include <bindings.dsl.h>
#define USE_SURFACE_BTREE
#include <stdio.h>
#include <gts.h>
#include "GTS_inline.h"

-- REMEMBER ForeignPtr for garbage collection!

-- | Bindings DSL file for the Gnu Triangulated Surface Library
module Bindings.GTS where 
#strict_import
-- import Bindings.GObject
import Bindings.GLib.Fundamentals
import Bindings.GLib.Fundamentals.BasicTypes

#cinline GTS_CHECK_VERSION, <guint> -> <guint> -> <guint> -> IO <gboolean>
#globalvar gts_allow_floating_vertices, IO <gboolean>

#starttype GSList
#field data, <gpointer>
#field next, Ptr <GSList>
#stoptype

-- | Allocate a node in a GSList from gobject. Note this function should not be used as the append/insert 
--    functions call it internally
#ccall g_slist_alloc, IO (Ptr <GSList>)

-- | Free all the nodes in an GSList. See the glib library 
#ccall g_slist_free, Ptr <GSList> -> IO ()

-- | Free one node in a GSList
#ccall g_slist_free_1, Ptr <GSList> -> IO ()

-- | Get the pointer to the next element in the GSList
#cinline g_slist_next, Ptr <GSList> -> IO (Ptr <GSList>)

-- | Get the pointer to the last element in the GSList
#ccall g_slist_last, Ptr <GSList> -> IO (Ptr <GSList>)

-- | Get the length of the GSList
#ccall g_slist_length, Ptr <GSList> -> IO <guint>

-- | Get the n-th element of the GSList counting from 0
#ccall g_slist_nth, Ptr <GSList> -> <guint> -> IO (Ptr <GSList>)

-- | Append a new node to a GSList 
#ccall g_slist_append, Ptr <GSList> -> <gpointer> -> IO (Ptr <GSList>)

-- | Insert a new node in a GSList before the referenced node 
#ccall g_slist_insert_before, Ptr <GSList> -> Ptr <GSList> -> <gpointer> -> IO (Ptr <GSList>)

-- | Insert a new node in a GSList at the specified position
#ccall g_slist_insert, Ptr <GSList> -> <gpointer> -> <gint> -> IO (Ptr <GSList>)

-- | Callback for the GTS Object class initialization - don't use this unless you know what you are doing
#callback GtsObjectClassInitFunc, Ptr <GtsObjectClass> -> IO ()

-- | Callback for the GTS Object initialization - don't use this unless you know what you are doing
#callback GtsObjectInitFunc, Ptr <GtsObject> -> IO ()

-- | Callback for the GTS argument set method - don't use this unless you know what you are doing
#callback GtsArgSetFunc, Ptr <GtsObject> -> IO ()

-- | Callback for the GTS argument get method - don't use this unless you know what you are doing
#callback GtsArgGetFunc, Ptr <GtsObject> -> IO ()

#num GTS_CLASS_NAME_LENGTH 

-- | Type of the GTS Vector object in Haskell
#starttype GtsVector
#stoptype

-- | Type of the GTS Matrix object in Haskell
#starttype GtsMatrix
#stoptype

-- | Create a new GTS Matrix 4x4 object and return the pointer to it.
#ccall gts_matrix_new,<gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> IO (Ptr <GtsMatrix>)

-- | Set the fields of an existing GTS Matrix object
#ccall gts_matrix_assign,Ptr <GtsMatrix> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> <gdouble> -> IO (Ptr <GtsMatrix>)

-- | Destroy a GTS matrix and free the associated memory - Use this with ForeignPtr for GC
#ccall gts_matrix_destroy,Ptr <GtsMatrix> -> IO ()

-- | Set a GTS matrix to the Zero matrix (if the matrix is NULL a new one is allocated)
#ccall gts_matrix_zero,Ptr <GtsMatrix> -> IO (Ptr <GtsMatrix>)

-- | Set a GTS matrix to the Identity matrix (if the matrix is NULL a new one is allocated)
#ccall gts_matrix_identity,Ptr <GtsMatrix> -> IO (Ptr <GtsMatrix>)

-- | Transpose a GTS Matrix and return the newly allocated matrix
#ccall gts_matrix_transpose,Ptr <GtsMatrix> -> IO (Ptr <GtsMatrix>)

-- | Invert a GTS Matrix and return the newly allocated matrix or NULL if the matrix can't be inverted
#ccall gts_matrix_inverse,Ptr <GtsMatrix> -> IO (Ptr <GtsMatrix>)

-- | Calculate the product of two matricies and return the newly allocated matrix
#ccall gts_matrix_product, Ptr <GtsMatrix> -> Ptr <GtsMatrix> -> IO (Ptr <GtsMatrix>)

-- | Scale a GTS Matrix in place 
#ccall gts_matrix_scale, Ptr <GtsMatrix> -> Ptr <GtsVector> -> IO (Ptr <GtsMatrix>)

-- | Translate the GTS Matrix by the GTS Vector (If the Matrix is NULL a new one is allocated and translated)
#ccall gts_matrix_translate, Ptr <GtsMatrix> -> Ptr <GtsVector> -> IO (Ptr <GtsMatrix>)

-- | Rotate the GTS Matrix around the vector by the 
#ccall gts_matrix_rotate, Ptr <GtsMatrix> -> Ptr <GtsVector> -> <gdouble> -> IO (Ptr <GtsMatrix>)

-- | Type for an (r,g,b) triple in GTS (floating point)
#starttype GtsColor
#field r, <gfloat>
#field g, <gfloat>
#field b, <gfloat>
#stoptype

-- | Callback for most GTS visitors 
#callback GtsFunc, <gpointer> -> <gpointer> -> IO <gint>

-- | Callback for gts_vertices_merge
#callback GtsVertexMergeFunc, Ptr <GtsVertex> -> Ptr <GtsVertex> -> IO <gboolean>

-- | GTS formatted data file 
#starttype GtsFile
#field line,<guint>
#field pos,<guint>
#field error,Ptr <gchar>
#stoptype

-- | A GTS formatted data file variable
#starttype GtsFileVariable
#stoptype

-- | Open a C file
#ccall fopen,CString -> CString -> IO (Ptr CFile)

-- | Close a C file
#ccall fclose, Ptr CFile -> IO Int

-- | Create a new GTS file handle from an C file
#ccall gts_file_new, Ptr CFile -> IO (Ptr <GtsFile>)

-- | Create a new GTS file handle from a C String. Note this method doesn't seem to work
#ccall gts_file_new_from_string, Ptr <gchar> -> IO ( Ptr <GtsFile>)

-- | Get a character from a GTS file (Don't use this)
#ccall gts_file_getc, Ptr <GtsFile> -> IO <gint>

-- | Destroy a GTS file handle and free the memory
#ccall gts_file_destroy, IO (Ptr <GtsFile>)

-- #cinline GTS_OBJECT_FLAGS, Ptr <GtsObject> -> <guint32>

#starttype GtsObjectClassInfo
#array_field name,<gchar>
#field object_size,<guint>
#field class_size,<guint>
#stoptype

#starttype GtsObjectClass
#field info,<GtsObjectClassInfo>
#field parent_class,Ptr <GtsObjectClass>
#stoptype

#starttype GtsObject
#field klass, Ptr <GtsObjectClass>
#field flags, <guint32>
#stoptype

#starttype GtsPointClass
#field parent_class, <GtsObjectClass>
#field binary, <gboolean>
#stoptype

#starttype GtsPoint
#field object, Ptr <GtsObject>
#field x,<gdouble>
#field y,<gdouble>
#field z,<gdouble>
#stoptype

#ccall gts_object_class, IO <gpointer>
#ccall gts_object_class_new, Ptr <GtsObjectClass> -> Ptr <GtsObjectClassInfo> -> IO <gpointer>
#ccall gts_object_check_cast, <gpointer> -> <gpointer> -> IO <gpointer>
#ccall gts_object_class_check_cast, <gpointer> -> <gpointer> -> IO <gpointer>
-- #cinline gts_object_is_from_class,  <gpointer> -> <gpointer> -> IO <gpointer>
-- #cinline gts_object_class_is_from_class,  <gpointer> -> <gpointer> -> IO <gpointer>
#ccall gts_object_class_from_name,Ptr <gchar> -> IO (Ptr <GtsObjectClass>)

#starttype GtsRange
#field min, <gdouble>
#field max, <gdouble>
#field sum, <gdouble>
#field sum2, <gdouble>
#field mean, <gdouble>
#field stddev, <gdouble>
#field n,<guint>
#stoptype

#starttype GtsSegment
#stoptype

#starttype GtsTriangleClass
#stoptype

#starttype GtsTriangle
#field e1, Ptr <GtsEdge>
#field e2, Ptr <GtsEdge>
#field e3, Ptr <GtsEdge>
#stoptype

-- | Create a triangle which is guarenteed to enclose all the points in the list
#ccall gts_triangle_enclosing, Ptr <GtsTriangleClass> -> Ptr <GSList> -> <gdouble> -> IO (Ptr <GtsTriangle>)

-- | Get the class descriptor for the GTS Triangle class
#ccall gts_triangle_class, IO (Ptr <GtsTriangleClass>)

-- | Create a new triangle from 3 edges
#ccall gts_triangle_new, Ptr <GtsTriangleClass> -> Ptr <GtsEdge> -> Ptr <GtsEdge> -> Ptr <GtsEdge> -> IO (Ptr <GtsTriangle>)

-- | Discard the existing edges of the triangle and replace with the new ones
#ccall gts_triangle_set, Ptr <GtsTriangleClass> -> Ptr <GtsEdge> -> Ptr <GtsEdge> -> Ptr <GtsEdge> -> IO ()

-- | Get the area of this triangle
#ccall gts_triangle_area, Ptr <GtsTriangle> -> IO <gdouble>

-- | Get the perimeter of this triangle
#ccall gts_triangle_perimeter, Ptr <GtsTriangle> -> IO <gdouble> 

-- | Get a measure of the quality of this triangle (how close to equilateral it is?)
#ccall gts_triangle_quality, Ptr <GtsTriangle> -> IO <gdouble>

-- | Get the normal to the plane of this triangle
#ccall gts_triangle_normal, Ptr <GtsTriangle> -> Ptr <gdouble> -> Ptr <gdouble> -> Ptr <gdouble> -> IO ()

-- | Changes the orientation of triangle t, turning it inside out
#ccall gts_triangle_revert, Ptr <GtsTriangle> -> IO ()

#ccall gts_triangle_orientation, Ptr <GtsTriangle> -> IO <gdouble>
#ccall gts_triangle_neighbors, Ptr <GtsTriangle> -> IO (Ptr <GSList>)
#ccall gts_triangle_vertices_edges, Ptr <GtsTriangle> -> Ptr <GtsEdge> -> Ptr (Ptr <GtsVertex>) -> Ptr (Ptr <GtsVertex>) -> Ptr (Ptr <GtsVertex>) -> Ptr (Ptr <GtsEdge>) -> Ptr (Ptr <GtsEdge>) -> Ptr (Ptr <GtsEdge>) -> IO ()
#ccall gts_triangle_vertices, Ptr <GtsTriangle> -> Ptr <GtsEdge> -> Ptr (Ptr <GtsVertex>) -> Ptr (Ptr <GtsVertex>) -> Ptr (Ptr <GtsVertex>) -> IO ()
#ccall gts_triangle_vertex_opposite, Ptr <GtsTriangle> -> Ptr <GtsEdge> -> IO (Ptr <GtsVertex)
#ccall gts_triangle_edge_opposite, Ptr <GtsTriangle> -> Ptr <GtsVertex> -> IO (Ptr <GtsEdge) 
#ccall gts_triangle_is_ok, Ptr <GtsTriangle> -> IO <gboolean>
#ccall gts_triangle_circumcircle_center, Ptr <GtsTriangle> -> Ptr <GtsPointClass> -> IO (Ptr <GtsPoint>)
#ccall gts_triangle_interpolate_height, Ptr <GtsTriangle> -> Ptr <GtsPoint> -> IO ()
#ccall gts_triangles_from_edges, Ptr <GSList> -> IO (Ptr <GSList>)

#starttype GtsVertex
#field p,<GtsPoint>
#stoptype

#starttype GtsVertexClass
#stoptype

-- | Clear a range 
#ccall gts_range_init, Ptr <GtsRange> -> IO ()
#ccall gts_range_reset,Ptr <GtsRange> -> IO () 
#ccall gts_range_add_value,Ptr <GtsRange> -> <gdouble> -> IO () 
#ccall gts_range_update,Ptr <GtsRange> -> IO () 

-- | GLib class pointer for the GTS point object class
#ccall gts_point_class, IO (Ptr <GtsPointClass>)

-- | Create a new GTS point in 3d space
#ccall gts_point_new, Ptr <GtsPointClass> -> <gdouble> -> <gdouble> -> <gdouble> -> IO ( Ptr <GtsPoint> )

-- | Set the value of a GTS point in 3d space
#ccall gts_point_set, Ptr <GtsPoint> -> <gdouble> -> <gdouble> -> <gdouble> -> IO ()

-- | True iff the point is within or on the boundary of the box defined by the two other points
#cinline gts_point_is_in_rectangle, Ptr <GtsPoint> -> Ptr <GtsPoint> -> Ptr <GtsPoint> -> IO <gboolean>

-- | GLib class pointer for the GTS vertex class (vertex derives from point)
#ccall gts_vertex_class,IO (Ptr <GtsVertexClass>)

-- | Create a new GTS vertex in 3d space
#ccall gts_vertex_new,Ptr <GtsVertexClass> -> <gdouble> -> <gdouble> -> <gdouble> -> IO (Ptr <GtsVertex>)

-- | True if this vertex is not part of a GTS segment
#ccall gts_vertex_is_unattached, Ptr <GtsVertex> -> IO <gboolean>

-- | Return the number of connected triangles sharing the vertex, if second parameter is true then sever the connection
#ccall gts_vertex_is_contact, Ptr <GtsVertex> -> <gboolean> -> IO <guint>

-- | Null unless two vertices are the endpoints of the same segment, in which case return the segment
#ccall gts_vertices_are_connected, Ptr <GtsVertex> -> Ptr <GtsVertex> -> IO (Ptr <GtsSegment)

-- | Replace a vertex with another vertex and update all objects using it
#ccall gts_vertex_replace, Ptr <GtsVertex> -> Ptr <GtsVertex> -> IO ()

-- | Adds to list all the GtsVertex connected to v by a GtsSegment and not already in list. If surface is not NULL only the vertices connected to v by an edge belonging to surface are considered.
#ccall gts_vertex_neighbors, Ptr <GtsVertex> -> Ptr <GSList> -> Ptr <GtsSurface> -> IO (Ptr <GSList>)

-- | Adds all the GtsTriangle which share v as a vertex and do not already belong to list.
#ccall gts_vertex_triangles, Ptr <GtsVertex> -> Ptr <GSList> -> IO (Ptr <GSList>)

-- | Adds all the GtsFace belonging to surface (if not NULL) which share v as a vertex and do not already belong to list.
#ccall gts_vertex_faces, Ptr <GtsVertex> -> Ptr <GtsSurface> -> Ptr <GSList> -> IO (Ptr <GSList>)

-- | A list of GtsEdge describing in counterclockwise order the boundary of the fan of summit v, the faces of the fan belonging to surface
#ccall gts_vertex_fan_oriented, Ptr <GtsVertex> -> Ptr <GtsSurface> -> IO (Ptr <GSList>)

-- | TRUE if v is strictly contained in the diametral circle of e, FALSE otherwise
#ccall gts_vertex_encroaches_edge, Ptr <GtsVertex> -> Ptr <GtsEdge> -> IO <gboolean>

-- | For each vertex v in vertices look if there are any vertex of vertices contained in a box centered on v of size 2*epsilon. If there are and if check is not NULL and returns TRUE, replace them with v (using gts_vertex_replace()), destroy them and remove them from list. This is done efficiently using Kd-Trees.
#ccall gts_vertices_merge, Ptr <GSList> -> <gdouble> -> <GtsVertexMergeFunc> -> IO (Ptr <GSList>)

#starttype GtsVertexNormal
#field n,<GtsVector>
#stoptype

#ccall gts_vertex_normal_class,IO (Ptr <GtsVertexClass>)

#starttype GtsColorVertex
#field c,<GtsColor>
#stoptype

#ccall gts_color_vertex_class,IO (Ptr <GtsVertexClass>)

#starttype GtsSurface
#field keep_faces,<gboolean>
#stoptype

#starttype GtsSurfaceClass
#stoptype

#starttype GtsEdgeClass
#stoptype

#starttype GtsEdge
#field triangles,Ptr <GSList>
#field segment,Ptr <GtsSegment>
#stoptype

#starttype GtsFaceClass
#stoptype

#starttype GtsFace
#stoptype

#starttype GtsSurfaceStats
#stoptype

#starttype GtsSurfaceQualityStats
#stoptype

#ccall gts_face_class,IO (Ptr <GtsFaceClass>)
#ccall gts_edge_class,IO (Ptr <GtsEdgeClass>)
#ccall gts_surface_class,IO (Ptr <GtsSurfaceClass>)

-- | Create a new empty surface which uses the specified types of sub-object
#ccall gts_surface_new,Ptr <GtsSurfaceClass> -> Ptr <GtsFaceClass> -> Ptr <GtsEdgeClass> -> Ptr <GtsVertexClass> -> IO (Ptr <GtsSurface>)

-- | Add a face to a surface
#ccall gts_surface_add_face, Ptr <GtsSurface> -> Ptr <GtsFace> -> IO ()

-- | Read a surface from a GTS formatted file
#ccall gts_surface_read, Ptr <GtsSurface> -> Ptr <GtsFile> -> IO <guint>

-- | Remove a face from a surface
#ccall gts_surface_remove_face,Ptr <GtsSurface> -> Ptr <GtsFace> -> IO ()

-- | Get the surface area of all triangles in the surface
#ccall gts_surface_area, Ptr <GtsSurface> -> IO <gdouble>

-- | Get some statistics on the surface
#ccall gts_surface_stats, Ptr <GtsSurface> -> Ptr <GtsSurfaceStats> -> IO ()

-- | Get some statistics on the quality of the triangles making up the surface
#ccall gts_surface_quality_stats, Ptr <GtsSurface> -> Ptr <GtsSurfaceQualityStats> -> IO ()

-- | Invoke a function for each vertex in the surface
#ccall gts_surface_foreach_vertex, Ptr <GtsSurface> -> <GtsFunc> -> <gpointer> -> IO ()

-- | Invoke a function for each edge in the surface
#ccall gts_surface_foreach_edge, Ptr <GtsSurface> -> <GtsFunc> -> <gpointer> -> IO ()

-- | Invoke a function for each face in the surface
#ccall gts_surface_foreach_face, Ptr <GtsSurface> -> <GtsFunc> -> <gpointer> -> IO ()

-- | Invoke a function for each face in the surface and remove the face afterwards
#ccall gts_surface_foreach_face_remove, Ptr <GtsSurface> -> <GtsFunc> -> <gpointer> -> IO <guint>

-- | Generate a surface which is a tessalated model of a sphere
#ccall gts_surface_generate_sphere, Ptr <GtsSurface> -> <guint> -> IO ( Ptr <GtsSurface> )

-- |Add a copy of all the faces, edges and vertices of s2 to s1. 
#ccall gts_surface_copy, Ptr <GtsSurface> -> Ptr <GtsSurface> -> IO (Ptr <GtsSurface> )

-- | Adds all the faces of with which do not already belong to s to s.
#ccall gts_surface_merge, Ptr <GtsSurface> -> Ptr <GtsSurface> -> IO ()

-- | True iff the surface describes a manifold
#ccall gts_surface_is_manifold, Ptr <GtsSurface> -> IO <gboolean>

-- | True iff the surface is closed
#ccall gts_surface_is_closed, Ptr <GtsSurface> -> IO <gboolean>

-- | True iff all the faces of the surface have a compatible orientation
#ccall gts_surface_is_orientable, Ptr <GtsSurface> -> IO <gboolean>

-- | Return the volume of the domain bounded by the surface, only valid if the surface is closed and orientable
#ccall gts_surface_volume, Ptr <GtsSurface> -> IO <gdouble>

-- | Return the center of mass of the domain bounded by the surface s, only valid if the surface is closed and orientable
#ccall gts_surface_center_of_mass, Ptr <GtsSurface> -> Ptr <GtsVector> -> IO <gdouble>

-- | Return the center of area of the surface (all faces should be co-planar)
#ccall gts_surface_center_of_area, Ptr <GtsSurface> -> Ptr <GtsVector> -> IO <gdouble>

#ccall gts_surface_vertex_number, Ptr <GtsSurface> -> IO <guint>
#ccall gts_surface_edge_number, Ptr <GtsSurface> -> IO <guint>
#ccall gts_surface_face_number, Ptr <GtsSurface> -> IO <guint>

#ccall gts_surface_boundary, Ptr <GtsSurface> -> IO (Ptr <GSList>)
#ccall gts_surface_split, Ptr <GtsSurface> -> IO (Ptr <GSList>)
#ccall gts_surface_write, Ptr <GtsSurface> -> Ptr CFile -> IO ()

#starttype GtsSurfaceInterClass
#stoptype

#starttype GtsSurfaceInter
#field s1,Ptr <GtsSurface>
#field s2,Ptr <GtsSurface>
#stoptype

#starttype GNode
#stoptype

-- | Create a new GTS Face from the 3 edges
#ccall gts_face_new, Ptr <GtsFaceClass> -> Ptr <GtsEdge> -> Ptr <GtsEdge> -> Ptr <GtsEdge> -> IO (Ptr <GtsFace>)

-- | Get the class for a surface intersection
#ccall gts_surface_inter_class,IO (Ptr <GtsSurfaceInterClass>)

-- | Create a new surface intersection from the two surfaces and the precomputed face bounding box trees
#ccall gts_surface_inter_new, Ptr <GtsSurfaceInterClass> -> Ptr <GtsSurface> -> Ptr <GtsSurface> -> Ptr <GNode> -> Ptr <GNode> -> <gboolean> -> <gboolean> -> IO (Ptr <GtsSurfaceInter>)

-- | True iff the edges in the intersection form a closed curve
#ccall gts_surface_inter_check, Ptr <GtsSurfaceInter> -> Ptr <gboolean> -> IO <gboolean>

-- | Adds to surface the part of the surface described by si and op.
#ccall gts_surface_inter_boolean, Ptr <GtsSurfaceInter> -> Ptr <GtsSurface> -> <guint> -> IO ()

-- | A new GtsSurface containing the faces of s which are self-intersecting or NULL if no faces of s are self-intersecting.
#ccall gts_surface_is_self_intersecting, Ptr <GtsSurface> -> IO (Ptr <GtsSurface>)

-- | A list of GtsEdge defining the curve intersection of the two surfaces.
#ccall gts_surface_intersection, Ptr <GtsSurface> -> Ptr <GtsSurface> -> Ptr <GNode> -> Ptr <GNode> -> IO (Ptr <GSList>)

#ccall gts_bb_tree_new, Ptr <GSList> -> IO (Ptr <GNode>)
#ccall gts_bb_tree_surface, Ptr <GtsSurface> -> IO (Ptr <GNode>)

#ccall gts_bb_tree_destroy, Ptr <GNode> -> <gboolean> -> IO ()

-- | TRUE if the ray starting at p and ending at (+infty, p->y, p->z) intersects with bb, FALSE otherwise.
#ccall gts_bb_tree_stabbed, Ptr <GNode> -> Ptr <GtsPoint> -> IO (Ptr <GSList>)

-- | Destroy a GTS object and free the associated memory. This is the generic version, see the typed helper functions
#ccall gts_object_destroy, Ptr <GtsObject> -> IO ()

-- #starttype GtsConstraintClass
-- #stoptype 

-- #starttype GtsConstraint
-- #stoptype

#callback GtsEncroachFunc, Ptr <GtsVertex> -> Ptr <GtsEdge> -> Ptr <GtsSurface> -> <gpointer> -> IO <gboolean>
#callback GtsKeyFunc, <gpointer> -> <gpointer> -> IO <gdouble>

#ccall gts_constraint_class, IO (Ptr <GtsEdgeClass>)
#ccall gts_point_locate, Ptr <GtsPoint> -> Ptr <GtsSurface> -> Ptr <GtsFace> -> IO (Ptr <GtsFace>)

-- | Create a new edge from a pair of vertices
#ccall gts_edge_new, Ptr <GtsEdgeClass> -> Ptr <GtsVertex> -> Ptr <GtsVertex> -> IO (Ptr <GtsEdge>)

-- | Replaces e with with. For each triangle which uses e as an edge, e is replaced with with. The with->triangles list is updated appropriately and the e->triangles list is freed and set to NULL
#ccall gts_edge_replace, Ptr <GtsEdge> -> Ptr <GtsEdge> -> IO ()

-- | Performs an "edge swap" on the two triangles sharing e and belonging to s.
#ccall gts_edge_swap, Ptr <GtsEdge> -> Ptr <GtsSurface> -> IO ()

#ccall gts_segment_triangle_intersection, Ptr <GtsSegment> -> Ptr <GtsTriangle> -> <gboolean> -> IO (Ptr <GtsPoint>)

-- | Tests if the planar projection (x, y) of p is inside or outside the circumcircle of the planar projection of t. This function is geometrically robust.
#ccall gts_point_in_triangle_circle,Ptr <GtsPoint> -> Ptr <GtsTriangle> -> IO <gdouble>

-- | Checks for orientation of the projection of three points on the (x,y) plane. The result is also an approximation of twice the signed area of the triangle defined by the three points. This function uses adaptive floating point arithmetic and is consequently geometrically robust.
#ccall gts_point_orientation, Ptr <GtsPoint> -> Ptr <GtsPoint> -> Ptr <GtsPoint> -> IO <gdouble>

-- | TRUE if p is inside the surface defined by tree, FALSE otherwise. 
#ccall gts_point_is_inside_surface, Ptr <GtsPoint> -> Ptr <GNode> -> <gboolean> -> IO <gboolean>

-- | Tests if the planar projection (x, y) of p is inside or outside the circle defined by the planar projection of p1, p2 and p3.
#ccall gts_point_in_circle, Ptr <GtsPoint> -> Ptr <GtsPoint> -> Ptr <GtsPoint> -> Ptr <GtsPoint> -> IO <gdouble>

-- | Checks if p4 lies above, below or on the plane passing through the points p1, p2 and p3. Below is defined so that p1, p2 and p3 appear in counterclockwise order when viewed from above the plane. The returned value is an approximation of six times the signed volume of the tetrahedron defined by the four points. This function uses adaptive floating point arithmetic and is consequently geometrically robust.
#ccall gts_point_orientation_3d, Ptr <GtsPoint> -> Ptr <GtsPoint> -> Ptr <GtsPoint> -> Ptr <GtsPoint> -> Ptr <GtsPoint> -> IO <gdouble>

#starttype GtsBBoxClass
#stoptype
#ccall gts_bbox_class, IO (Ptr <GtsBBoxClass>)

#starttype GtsBBox
#field x1,<gdouble>
#field y1,<gdouble>
#field z1,<gdouble>
#field x2,<gdouble>
#field y2,<gdouble>
#field z2,<gdouble>
#field bounded, <gpointer>
#stoptype

-- | A list of triangle strips containing all the triangles of s. A triangle strip is itself a list of successive triangles having one edge in common.
#ccall gts_surface_strip, Ptr <GtsSurface> -> IO (Ptr <GSList>)

-- | Using the gts_bb_tree_surface_distance() and gts_bb_tree_surface_boundary_distance() functions fills face_range and boundary_range with the min, max and average Euclidean (minimum) distances between the faces of s1 and the faces of s2 and between the boundary edges of s1 and s2.
#ccall gts_surface_distance, Ptr <GtsSurface> -> Ptr <GtsSurface> -> <gdouble> -> Ptr <GtsRange> -> Ptr <GtsRange> -> IO ()

-- | A new GtsBBox bounding box of surface.
#ccall gts_bbox_surface, Ptr <GtsBBoxClass> -> Ptr <GtsSurface> -> IO ( Ptr <GtsBBox> )

-- | TRUE if the bounding boxes bb1 and bb2 are overlapping (including just touching), FALSE otherwise.
#ccall gts_bboxes_are_overlapping, Ptr <GtsBBox> -> Ptr <GtsBBox> -> IO <gboolean>

-- | Add a constraint edge to a Delaunay surface
#ccall gts_delaunay_add_constraint, Ptr <GtsSurface> -> Ptr <GtsEdge> -> IO (Ptr <GSList>)

-- | NULL if the planar projection of surface is a Delaunay triangulation (unconstrained), a GtsFace violating the Delaunay property otherwise.
#ccall gts_delaunay_check, Ptr <GtsSurface> -> IO (Ptr <GtsFace>)

-- | Removes all the edges of the boundary of surface which are not constraints.
#ccall gts_delaunay_remove_hull, Ptr <GtsSurface> -> IO ()

-- | Recursively split constraints of surface which are encroached by vertices of surface (see Shewchuk 96 for details). The split constraints are destroyed and replaced by a set of new constraints of the same class. If gts_vertex_encroaches_edge() is used for encroaches, the resulting surface will be Delaunay conforming. If steiner_max is positive or nul, the recursive splitting procedure will stop when this maximum number of Steiner points is reached. In that case the resulting surface will not necessarily be Delaunay conforming.
#ccall gts_delaunay_conform,Ptr <GtsSurface> -> <gint> -> FunPtr <GtsEncroachFunc> -> <gpointer> -> IO <guint>

-- | An implementation of the refinement algorithm described in Ruppert (1995) and Shewchuk (1996).
#ccall gts_delaunay_refine,Ptr <GtsSurface> -> <gint> -> FunPtr <GtsEncroachFunc> -> <gpointer> -> FunPtr <GtsKeyFunc> -> <gpointer> -> IO <guint>

-- | Add one vertex to a Delaunay triangulation preserving the Delaunay property
#ccall gts_delaunay_add_vertex, Ptr <GtsSurface> -> Ptr <GtsVertex> -> Ptr <GtsFace> -> IO (Ptr <GtsVertex>)

-- | Add one vertex to a face of a Delaunay triangulation preserving the Delaunay property
#ccall gts_delaunay_add_vertex_to_face, Ptr <GtsSurface> -> Ptr <GtsVertex> -> Ptr <GtsFace> -> IO (Ptr <GtsVertex>)

-- | Removes v from the Delaunay triangulation defined by surface and restores the Delaunay property. Vertex v must not be used by any constrained edge otherwise the triangulation is not guaranteed to be Delaunay.
#ccall gts_delaunay_remove_vertex, Ptr <GtsSurface> -> Ptr <GtsVertex> -> IO ()

-- Typed helper functions
#ccall gts_surface_destroy, Ptr <GtsSurface> -> IO ()
#ccall gts_triangle_destroy, Ptr <GtsTriangle> -> IO ()
#ccall gts_vertex_destroy, Ptr <GtsVertex> -> IO ()